@@ -57,6 +57,12 @@ class ActionRules:
         Decisions object.
     is_strict_flexible : bool
         If true flexible attributes must be always actionable, if false they can also behave as stable attributes
+    util : List[pd.Series] = None
+        List of utilities for classification rules.
+    min_util_dif : float = None
+        Number representing minimal desired change in utility caused by action.
+    sort_by_util_dif : bool = False
+        Should the output action rules be sorted by utility difference?
 
     Methods
     -------
@@ -82,7 +88,12 @@ class ActionRules:
                  min_flexible_antecedents: int = 1,
                  max_stable_antecedents: int = 1,
                  max_flexible_antecedents: int = 1,
-                 is_strict_flexible: bool = True
+                 is_strict_flexible: bool = True,
+                 util_flex: List[pd.Series] = None,
+                 util_target: List[pd.Series] = None,
+                 min_util_dif: float = None,
+                 min_profit: float = None,
+                 sort_by_util_dif: bool = False
                  ):
         """
         Parameters
@@ -113,6 +124,14 @@ class ActionRules:
             Maximal number of flexible pairs.
         is_strict_flexible : bool
             If true flexible attributes must be always actionable, if false they can also behave as stable attributes
+        util : List[pd.Series] = None
+            List of utilities for classification rules.
+        min_util_dif : float = None
+            Number representing minimal desired change in utility caused by action.
+        min_profit : float = None
+            Number representing minimal profit.
+        sort_by_util_dif : bool = False
+            Should the output action rules be sorted by utility difference?
         """
         self.stable_tables = stable_tables
         self.flexible_tables = flexible_tables
@@ -135,6 +154,11 @@ class ActionRules:
         self.not_default_target_classes = self.desired_state.get_not_in_default_classes()
         self.decisions = decisions
         self.is_strict_flexible = is_strict_flexible
+        self.util_flex = util_flex
+        self.util_target = util_target
+        self.min_util_dif = min_util_dif
+        self.min_profit = min_profit
+        self.sort_by_util_dif = sort_by_util_dif
 
     def _is_action_couple(self,
                           before: Union[str, int, float],
@@ -225,7 +249,7 @@ class ActionRules:
                 count_antecedent += 1
                 action_rule_part.append([column, action_couple])
             elif action_couple is not None:
-                    action_rule_part.append([column, action_couple])
+                action_rule_part.append([column, action_couple])
         return True, action_rule_part, count_antecedent
 
     def _add_action_rule(self,
@@ -234,7 +258,9 @@ class ActionRules:
                          action_rule_decision: list,
                          action_rule_supp: list,
                          action_rule_conf: list,
-                         uplift: float):
+                         uplift: float,
+                         util_dif: float = None,
+                         profit: float = None):
         """This method joins the parts of an action rule and adds the action rule to a list.
 
         Parameters
@@ -251,9 +277,13 @@ class ActionRules:
             List of confidences.
         uplift: float
             Uplift
+        util_dif : float
+            Utility difference
+        profit : float
+            Profit
         """
         action_rule = [action_rule_stable, action_rule_flexible, action_rule_decision]
-        self.action_rules.append([action_rule, action_rule_supp, action_rule_conf, uplift])
+        self.action_rules.append([action_rule, action_rule_supp, action_rule_conf, uplift, util_dif, profit])
 
     def _split_to_before_after_consequent(self, decision_column: pd.DataFrame) -> tuple:
         """This method split the table based on consequent.
@@ -290,6 +320,14 @@ class ActionRules:
             supp = supp.astype(float)
             conf = self.conf.pop(0)
             conf = conf.astype(float)
+            util_flex = None
+            if self.util_flex and (self.min_util_dif is not None or self.min_profit is not None):
+                util_flex = self.util_flex.pop(0)
+                util_flex = util_flex.astype(float)
+            util_target = None
+            if self.util_target and (self.min_util_dif is not None or self.min_profit is not None):
+                util_target = self.util_target.pop(0)
+                util_target = util_target.astype(float)
             (before_indexes, after_indexes) = self._split_to_before_after_consequent(decision_column)
             for comb in itertools.product(before_indexes, after_indexes):
                 # Check if it is not used twice - just for reduction by nan
@@ -299,6 +337,33 @@ class ActionRules:
                     self.used_indexes.append(comb)
                 rule_before_index = comb[0]
                 rule_after_index = comb[1]
+
+                util_dif = None
+                profit = None
+                if util_flex is not None and util_target is not None:
+                    # utility difference check
+                    if self.min_util_dif is not None:
+                        utility_before = util_flex[rule_before_index] + util_target[rule_before_index]
+                        utility_after = util_flex[rule_after_index] + util_target[rule_after_index]
+                        util_dif = utility_after - utility_before
+                        if self.min_util_dif > util_dif:
+                            continue
+                    # utility profit check
+                    if self.min_profit is not None:
+                        util_flex_before = util_flex[rule_before_index]
+                        util_flex_after = util_flex[rule_after_index]
+                        util_target_before = util_target[rule_before_index]
+                        util_target_after = util_target[rule_after_index]
+                        cost = util_flex_after - util_flex_before
+                        conf_before = conf[rule_before_index]
+                        conf_after = conf[rule_after_index]
+                        benefit = (util_target_after - util_target_before) * (conf_after + conf_before - 1)
+                        # utility values for cost part can be either negative o positive - use of absolute values
+                        # makes it handles it the same
+                        profit = benefit - abs(cost)
+                        if self.min_profit > profit:
+                            continue
+
                 decision_before = decision_column.at[rule_before_index, decision_column.columns[0]]
                 decision_after = decision_column.at[rule_after_index, decision_column.columns[0]]
                 if self.desired_state.is_candidate_decision(decision_before, decision_after):
@@ -319,9 +384,9 @@ class ActionRules:
                     action_rule_decision = [
                         decision_column.columns[0], [decision_before, decision_after]]
                     if counted_flexible >= self.min_flexible_antecedents and \
-                       counted_stable >= self.min_stable_antecedents and \
-                       counted_flexible <= self.max_flexible_antecedents and \
-                       counted_stable <= self.max_stable_antecedents:
+                            counted_stable >= self.min_stable_antecedents and \
+                            counted_flexible <= self.max_flexible_antecedents and \
+                            counted_stable <= self.max_stable_antecedents:
                         if not self.is_nan:
                             support = min(supp[rule_before_index], supp[rule_after_index])
                             confidence = conf[rule_before_index] * conf[rule_after_index]
@@ -337,11 +402,12 @@ class ActionRules:
                                 confidence = None
                                 uplift = None
                             else:
-                                (left_support_before, support_before) = self._get_frequency_from_mask(action_rule_stable,
-                                                                                                      action_rule_flexible,
-                                                                                                      action_rule_decision,
-                                                                                                      0
-                                                                                                      )
+                                (left_support_before, support_before) = self._get_frequency_from_mask(
+                                    action_rule_stable,
+                                    action_rule_flexible,
+                                    action_rule_decision,
+                                    0
+                                    )
                                 (left_support_after, support_after) = self._get_frequency_from_mask(action_rule_stable,
                                                                                                     action_rule_flexible,
                                                                                                     action_rule_decision,
@@ -349,7 +415,8 @@ class ActionRules:
                                                                                                     )
                                 support = support_before / total
                                 if left_support_before != 0 and left_support_after != 0:
-                                    confidence = (support_before / left_support_before) * (support_after / left_support_after)
+                                    confidence = (support_before / left_support_before) * (
+                                                support_after / left_support_after)
                                     uplift = self._get_uplift(
                                         support_before,
                                         (support_before / left_support_before),
@@ -371,10 +438,16 @@ class ActionRules:
                                               action_rule_decision,
                                               action_rule_supp,
                                               action_rule_conf,
-                                              uplift)
+                                              uplift,
+                                              util_dif,
+                                              profit)
                         self.classification_before.append(rule_before_index)
                         self.classification_after.append(rule_after_index)
 
+        # sort by utility difference
+        if util_flex is not None and util_target is not None and self.sort_by_util_dif:
+            self.action_rules.sort(key=lambda rule: rule[4], reverse=True)
+
     def pretty_text(self):
         """It generates human language representation of action rules.
 
@@ -384,6 +457,8 @@ class ActionRules:
             supp = row[1]
             conf = row[2]
             uplift = row[3]
+            util_dif = row[4]
+            profit = row[5]
             text = "If "
             # Stable part
             stable_part = action_rule[0]
@@ -402,7 +477,11 @@ class ActionRules:
             decision = action_rule[2]
             text += "then '" + str(decision[0]) + "' value '" + str(decision[1][0]) + "' is changed to '" + \
                     str(decision[1][1]) + "' with support: " + str(supp[2]) + ", confidence: " + str(conf[2]) + \
-            " and uplift: " + str(uplift) + "."
+                    " and uplift: " + str(uplift) + "."
+            if util_dif is not None:  # action rule contains utility
+                text += " Change in utility caused by action is " + str(util_dif) + "."
+            if profit is not None:  # action rule contains utility
+                text += " Profit of the action is " + str(profit) + "."
             self.action_rules_pretty_text.append(text)
 
     def representation(self):
@@ -414,6 +493,8 @@ class ActionRules:
             supp = row[1]
             conf = row[2]
             uplift = row[3]
+            util_dif = row[4]
+            profit = row[5]
             text = "r = [   "
             # Stable part
             stable_part = action_rule[0]
@@ -433,7 +514,12 @@ class ActionRules:
             decision = action_rule[2]
             text += "] ⇒ [" + str(decision[0]) + ": " + str(decision[1][0]) + " → " + \
                     str(decision[1][1]) + "] with support: " + str(supp[2]) + ", confidence: " + str(
-                conf[2]) + " and uplift: " + str(uplift) + "."
+                conf[2]) + ", uplift: " + str(uplift)
+            if util_dif is not None:  # action rule contains utility
+                text += ", change in utility: " + str(util_dif)
+            if profit is not None:
+                text += ", profit is " + str(profit)
+            text += "."
             self.action_rules_representation.append(text)
 
     @staticmethod
@@ -461,7 +547,8 @@ class ActionRules:
         else:
             return 0
 
-    def _get_frequency_from_mask(self, action_rule_stable: list, action_rule_flexible: list, action_rule_decision: list, part: int) -> tuple:
+    def _get_frequency_from_mask(self, action_rule_stable: list, action_rule_flexible: list, action_rule_decision: list,
+                                 part: int) -> tuple:
         """Get frequency from source data.
 
         Parameters
@@ -510,4 +597,4 @@ class ActionRules:
 
         support = len(new_data)
 
-        return left_support, support
\ No newline at end of file
+        return left_support, support
